# https://html.spec.whatwg.org/multipage/server-sent-events.html
from __future__ import annotations

import sys
import time
from email.message import Message
from functools import wraps
from io import StringIO
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    Generic,
    NamedTuple,
    TypeVar,
)
from weakref import proxy

import anyio
import httpx

if TYPE_CHECKING:
    from typing import TypedDict
    from typing_extensions import Never, Required, Unpack

    from httpx._types import (
        AuthTypes,
        CookieTypes,
        HeaderTypes,
        QueryParamTypes,
        TimeoutTypes,
        URLTypes,
    )

    class _RequestKwargs(TypedDict, total=False):
        url: Required[URLTypes]
        params: QueryParamTypes
        headers: HeaderTypes
        cookies: CookieTypes
        auth: AuthTypes
        follow_redirects: bool
        timeout: TimeoutTypes
        extensions: dict


_ClientT = TypeVar("_ClientT", httpx.Client, httpx.AsyncClient)

if sys.version_info < (3, 9):

    def _removeprefix(__string: str, __prefix: str) -> str:
        if __string.startswith(__prefix):
            return __string[len(__prefix) :]
        return __string
else:
    _removeprefix = str.removeprefix


def _pass_by_proxy(func):
    @wraps(func)
    def inner(self, *args, **kwargs):
        self = proxy(self)
        return func(self, *args, **kwargs)

    return inner


def _is_event_stream(response: httpx.Response):
    message = Message()
    content_type = response.headers.get("Content-Type")
    if content_type:
        message["Content-Type"] = content_type
    return message.get_content_type() == "text/event-stream"


class InvalidStateError(Exception):
    pass


class _InvalidState:
    def __bool__(self):
        return False


InvalidState: Any = _InvalidState()


class Event(NamedTuple):
    type: str
    data: str
    origin: httpx.URL
    last_event_id: str


class EventSource(Generic[_ClientT]):
    __slots__ = (
        "_client",
        "_request",
        "_origin",
        "_last_event_id",
        "_reconnect",
        "_inner",
        "__weakref__",
    )

    def __init__(self, client: _ClientT, **request: Unpack[_RequestKwargs]) -> None:
        self._client = client
        self._request = request
        timeout = request.pop("timeout", client.timeout)
        try:
            timeout = request["extensions"].pop("timeout")  # type: ignore (reportTypedDictNotRequiredAccess)
        except KeyError:
            pass
        timeout = httpx.Timeout(timeout)
        timeout.read = None
        request["timeout"] = timeout
        self._last_event_id = ""
        self._reconnect = 0
        self._inner = self.__inner()

    @property
    def last_event_id(self):
        return self._last_event_id

    @_pass_by_proxy
    def __inner(self) -> Generator[Event | None, str, Never]:
        event: Event | None = None
        last_event_id = ""
        event_type = ""

        with StringIO() as data_buffer:
            while True:
                line, event = (yield event), None
                if line is None:
                    event = InvalidState
                    continue
                field, colon, value = line.partition(":")
                del line
                value = _removeprefix(value, " ")

                if (field, colon) == ("", ""):
                    self._last_event_id = last_event_id
                    data = data_buffer.getvalue()
                    if not data:
                        event_type = ""
                        continue
                    event = Event(
                        type=event_type or "message",
                        data=data[:-1],
                        origin=self._origin,
                        last_event_id=last_event_id,
                    )
                    del data
                    data_buffer.seek(0)
                    data_buffer.truncate(0)
                    event_type = ""

                elif (field, colon) == ("", ":"):
                    continue

                elif field == "event":
                    event_type = value

                elif field == "data":
                    print(value, file=data_buffer)

                elif field == "id":
                    if value != "\x00":
                        last_event_id = value

                elif field == "retry":
                    try:
                        self._reconnect = int(value)
                    except ValueError:
                        pass

    def __iter__(self: EventSource[httpx.Client]):
        client = self._client
        request_args = self._request
        send_args = {
            "auth": request_args.pop("auth", httpx.USE_CLIENT_DEFAULT),
            "follow_redirects": request_args.pop(
                "follow_redirects", httpx.USE_CLIENT_DEFAULT
            ),
            "stream": True,
        }
        driver = self._inner
        try:
            if next(driver) is InvalidState:
                raise InvalidStateError(
                    f"{self.__class__.__name__} is already running!"
                )
        except StopIteration:
            raise InvalidStateError(
                f"{self.__class__.__name__} was already closed!"
            ) from None
        try:
            while True:
                request: httpx.Request = client.build_request("GET", **request_args)  # type: ignore (reportCallIssue)
                if self._last_event_id:
                    request.headers["Last-Event-ID"] = self._last_event_id

                time.sleep(self._reconnect)

                response = client.send(request, **send_args)
                try:
                    self._origin = response.url
                    if response.status_code == 204:
                        return
                    response.raise_for_status()
                    assert _is_event_stream(response)

                    yield from filter(None, map(driver.send, response.iter_lines()))

                except httpx.RemoteProtocolError:
                    continue
                finally:
                    response.close()
        finally:
            driver.close()

    async def __aiter__(self: EventSource[httpx.AsyncClient]):
        client = self._client
        request_args = self._request
        send_args = {
            "auth": request_args.pop("auth", httpx.USE_CLIENT_DEFAULT),
            "follow_redirects": request_args.pop(
                "follow_redirects", httpx.USE_CLIENT_DEFAULT
            ),
            "stream": True,
        }
        driver = self._inner
        try:
            if next(driver) is InvalidState:
                raise InvalidStateError(
                    f"{self.__class__.__name__} is already running!"
                )
        except StopIteration:
            raise InvalidStateError(
                f"{self.__class__.__name__} was already closed!"
            ) from None
        try:
            while True:
                request: httpx.Request = client.build_request("GET", **request_args)  # type: ignore (reportCallIssue)
                if self._last_event_id:
                    request.headers["Last-Event-ID"] = self._last_event_id

                await anyio.sleep(self._reconnect)

                response = await client.send(request, **send_args)
                try:
                    self._origin = response.url
                    if response.status_code == 204:
                        return
                    response.raise_for_status()
                    assert _is_event_stream(response)

                    async for line in response.aiter_lines():
                        event = driver.send(line)
                        if event:
                            yield event

                except StopIteration as exc:
                    raise RuntimeError(
                        f"{self.__class__.__name__} was already closed!"
                    ) from exc
                except httpx.RemoteProtocolError:
                    continue
                except Exception as exc:
                    driver.throw(exc)
                    return
                finally:
                    await response.aclose()
        finally:
            driver.close()

    def close(self):
        self._inner.close()



from __future__ import annotations

from typing import overload

import httpx

from ._limiter import RateLimiter
from ._url import API_URL

try:
    # guard against reloading
    _PINS  # type: ignore  # noqa: B018
except NameError:
    _PINS: dict[str, str] = {}


__all__ = ["NSAuth"]


class NSAuth(RateLimiter):
    @overload
    def __init__(self, *, password: str) -> None: ...

    @overload
    def __init__(self, *, autologin: str) -> None: ...

    def __init__(
        self, *, password: str | None = None, autologin: str | None = None
    ) -> None:
        self._password = password
        self.autologin = autologin

    @property
    def password(self) -> str | None:
        return self._password

    @password.setter
    def password(self, value: str | None) -> None:
        if value:
            self._autologin = None  # likely out of date
        self._password = value

    @property
    def autologin(self) -> str | None:
        return self._autologin

    @autologin.setter
    def autologin(self, value: str | None) -> None:
        if value:
            self._password = None  # no longer needed
        self._autologin = value

    def _request_hook(self, request: httpx.Request) -> httpx.Request:
        params = request.url.params
        nation = params.get("nation")
        if not nation:
            return request
        headers = request.headers
        if self._autologin:
            headers["X-Autologin"] = self._autologin
            pin = _PINS.get(self._autologin)
            if pin:
                headers["X-Pin"] = pin
        elif self._password:
            headers["X-Password"] = self._password
        try:
            is_empty = not request.content
        except httpx.RequestNotRead:
            pass  # don't override content
        else:
            if is_empty and ("c" in params or request.method == "POST"):
                # coerce a POST request
                headers.pop("Content-Type", None)
                headers.pop("Content-Length", None)
                request = httpx.Request(
                    "POST", API_URL, headers=headers, data=dict(params)
                )
        return super()._request_hook(request)

    def _response_hook(self, response: httpx.Response) -> None:
        autologin: str | None = response.headers.get("X-Autologin")
        if autologin:
            self.autologin = autologin
            pin = response.headers.get("X-Pin")
            if pin:
                _PINS[autologin] = pin
        return super()._response_hook(response)



from __future__ import annotations

import codecs
import zlib
from typing import Iterable, TypeVar
from xml.etree.ElementTree import Element, XMLParser, XMLPullParser

_T = TypeVar("_T")


class NullIncrementalDecoder:
    @staticmethod
    def decode(input: _T, final: bool = False) -> _T:
        return input


class GZipDecoder:
    def __init__(self) -> None:
        self._decompressobj = zlib.decompressobj(zlib.MAX_WBITS | 16)

    def decode(self, data: bytes) -> bytes:
        return self._decompressobj.decompress(data)

    def flush(self) -> bytes:
        return self._decompressobj.flush()


class XMLDecoder:
    def __init__(self, encoding: str | None = None) -> None:
        self._parser = XMLParser(encoding=encoding)

    def decode(self, data: bytes) -> None:
        self._parser.feed(data)

    def flush(self) -> Element:
        return self._parser.close()


class XMLChunker:
    def __init__(self, *, encoding: str | None = None) -> None:
        self._decoder = (
            codecs.getincrementaldecoder(encoding)("replace")
            if encoding
            else NullIncrementalDecoder
        )
        self._pull_parser = XMLPullParser(["start", "end"])
        self._path: list[Element] = []

    def decode(self, data: bytes) -> Iterable[Element]:
        self._pull_parser.feed(self._decoder.decode(data, False))
        return self._read_events()

    def flush(self) -> Iterable[Element]:
        self._pull_parser.feed(self._decoder.decode(b"", True))
        self._pull_parser.close()
        return self._read_events()

    def _read_events(self) -> Iterable[Element]:
        element: Element
        path = self._path
        for event, element in self._pull_parser.read_events():  # type: ignore
            if event == "start":
                path.append(element)
            elif event == "end":
                path.pop()
                if len(path) == 1:
                    yield element
                    path[0].clear()


try:
    from lxml.etree import (
        XMLParser as LXMLParser,
        _Element as LElement,
    )
    from lxml.objectify import ObjectifiedElement, ObjectifyElementClassLookup
except ImportError:
    pass
else:

    class LXMLDecoder:
        def __init__(self, encoding: str | None = None) -> None:
            self._parser = LXMLParser(encoding=encoding)

        def decode(self, data: bytes) -> None:
            self._parser.feed(data)

        def flush(self) -> LElement:
            return self._parser.close()

    class ObjectifyDecoder:
        def __init__(self, encoding: str | None = None) -> None:
            self._parser = parser = LXMLParser(
                encoding=encoding, remove_blank_text=True
            )
            parser.set_element_class_lookup(ObjectifyElementClassLookup())

        def decode(self, data: bytes) -> None:
            self._parser.feed(data)

        def flush(self) -> ObjectifiedElement:
            return self._parser.close()  # type: ignore



from __future__ import annotations

from email.message import Message
from itertools import chain
from typing import TYPE_CHECKING, Any, AsyncIterator, Iterator
from xml.etree.ElementTree import Element

import httpx

from ._decoder import GZipDecoder, XMLChunker, XMLDecoder
from ._errors import narrow

if TYPE_CHECKING:
    from typing_extensions import Self

    import xmltodict
    from lxml.etree import _Element
    from lxml.objectify import ObjectifiedElement

    from ._decoder import LXMLDecoder, ObjectifyDecoder

    HAS_XMLTODICT: bool = True
    HAS_LXML: bool = True
else:
    try:
        import xmltodict

        HAS_XMLTODICT = True
    except ModuleNotFoundError:
        HAS_XMLTODICT = False

    try:
        from lxml.etree import _Element
        from lxml.objectify import ObjectifiedElement

        HAS_LXML = True
    except ModuleNotFoundError:
        HAS_LXML = False
    else:
        from ._decoder import LXMLDecoder, ObjectifyDecoder

__all__ = ["Response"]


class Response(httpx.Response):
    @property
    def content_type(self) -> str:
        if not hasattr(self, "_content_type"):
            message = Message()
            content_type = self.headers.get("Content-Type")
            if content_type:
                message["Content-Type"] = content_type
            self._content_type = message.get_content_type()
        return self._content_type

    def iter_gzip(self) -> Iterator[bytes]:
        decoder = GZipDecoder()
        yield from map(decoder.decode, self.iter_bytes())
        yield decoder.flush()

    async def aiter_gzip(self) -> AsyncIterator[bytes]:
        decoder = GZipDecoder()
        async for chunk in self.aiter_bytes():
            yield decoder.decode(chunk)
        yield decoder.flush()

    if HAS_XMLTODICT:

        def json(self, **kwargs: Any) -> dict[str, Any]:
            if self.content_type.endswith("/xml"):
                return xmltodict.parse(self.content, encoding=self.encoding, **kwargs)
            return super().json(**kwargs)

    @property
    def xml(self) -> Element:
        if not hasattr(self, "_xml"):
            content = self.content
            decoder = XMLDecoder(self.encoding)
            decoder.decode(content)
            self._xml = decoder.flush()
        return self._xml

    def iter_xml(self) -> Iterator[Element]:
        decoder = XMLChunker(encoding=self.encoding)
        chunker = (
            self.iter_gzip()
            if self.content_type.endswith(("/x-gzip", "/gzip"))
            else self.iter_bytes()
        )
        yield from chain.from_iterable(map(decoder.decode, chunker))
        yield from decoder.flush()

    async def aiter_xml(self) -> AsyncIterator[Element]:
        decoder = XMLChunker(encoding=self.encoding)
        chunker = (
            self.aiter_gzip()
            if self.content_type.endswith(("/x-gzip", "/gzip"))
            else self.aiter_bytes()
        )
        async for chunk in chunker:
            for element in decoder.decode(chunk):
                yield element
        for element in decoder.flush():
            yield element

    if HAS_LXML:

        @property
        def lxml(self) -> _Element:
            if not hasattr(self, "_lxml"):
                content = self.content
                decoder = LXMLDecoder(self.encoding)
                decoder.decode(content)
                self._lxml = decoder.flush()
            return self._lxml

        @property
        def objectified(self) -> ObjectifiedElement:
            if not hasattr(self, "_objectified"):
                content = self.content
                decoder = ObjectifyDecoder(self.encoding)
                decoder.decode(content)
                self._objectified = decoder.flush()
            return self._objectified

    def raise_for_status(self) -> Self:  # type: ignore
        try:
            super().raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise narrow(exc).with_traceback(exc.__traceback__) from None
        return self



from __future__ import annotations

import logging
import time
import warnings
from contextlib import AsyncExitStack, ExitStack
from itertools import count, repeat, starmap
from math import ldexp
from random import random
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    ClassVar,
    Generator,
    Iterator,
    Mapping,
    TypeVar,
)

import anyio
import httpx

from ._errors import AgentNotSetError
from ._lock import ResetLock
from ._response import Response
from ._url import API_URL

if TYPE_CHECKING:
    pass

__all__ = ["RateLimiter", "TelegramLimiter"]
LOG = logging.getLogger("sans")
_KT = TypeVar("_KT")
_T = TypeVar("_T")


def _get_as_int(mapping: Mapping[_KT, Any], key: _KT, default: _T) -> int | _T:
    try:
        return int(mapping[key])
    except (KeyError, ValueError):
        return default


class _Backoff:
    def __getitem__(self, item: slice) -> Iterator[float]:
        if item.stop is None or item.step == 0:
            indices = count(item.start, item.step)
        else:
            indices = range(item.start or 0, item.stop, item.step or 1)

        # Yields random numbers in the range of [0, 2**index)
        return map(ldexp, starmap(random, repeat(())), indices)


_backoff = _Backoff()


def _raw_headers(headers: httpx.Headers) -> list[tuple[str, str]]:
    encoding = headers.encoding
    return [
        (key.decode(encoding, "replace"), value.decode(encoding, "replace"))
        for key, value in headers.raw
    ]


# TODO: refactor I/O flows into multiple functions
class RateLimiter(httpx.Auth):
    """
    httpx Auth utility object which implements ratelimiting, User-Agent management,
    and exponential backoff, while adding XML support to the resulting Response.
    """

    _agent: ClassVar = ""
    _lock: ClassVar = ResetLock()

    def _request_hook(self, request: httpx.Request) -> httpx.Request:
        agent = RateLimiter._agent
        if not agent:
            raise AgentNotSetError("sans has no set agent")
        request.headers["User-Agent"] = agent
        if LOG.isEnabledFor(logging.DEBUG):
            log = [
                f" > {request.method} {request.url.raw_path.decode('ascii', 'replace')} HTTP/1.1"
            ]
            for key, value in _raw_headers(request.headers):
                log.append(f"{key}: {value}")
            LOG.debug("\n > ".join(log))
        return request

    def _response_hook(self, response: httpx.Response) -> None:
        response_headers = response.headers
        if LOG.isEnabledFor(logging.INFO):
            log = [
                f" < {response.request.method} {response.url.raw_path.decode('ascii', 'replace')} HTTP/1.1"
            ]
            if LOG.isEnabledFor(logging.DEBUG):
                for key, value in _raw_headers(response_headers):
                    log.append(f"{key}: {value}")
                level = logging.DEBUG
            else:
                level = logging.INFO
            LOG.log(level, "\n < ".join(log))
        # Retry-After is handled before this point
        xrlr = _get_as_int(response_headers, "RateLimit-Remaining", 1)
        if xrlr:
            return
        xrlr = _get_as_int(response_headers, "RateLimit-Reset", 0)
        RateLimiter._lock.defer(xrlr)

    def auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response, None]:
        if request.url.host == API_URL.host:
            request = self._request_hook(request)
        response = yield request
        response.__class__ = Response
        if response.url.host == API_URL.host:
            self._response_hook(response)

    def sync_auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response, None]:
        recruitment_delay: int = getattr(self, "_recruitment_delay", 0)
        is_telegram_limiter = recruitment_delay > 0
        with ExitStack() as lock_stack:
            if request.headers["Host"] == API_URL.netloc.decode(
                request.headers.encoding
            ):
                if is_telegram_limiter:
                    lock_stack.enter_context(TelegramLimiter._lock)
                    if TelegramLimiter._last_request is not None:
                        time.sleep(
                            max(
                                TelegramLimiter._last_request
                                + recruitment_delay
                                - time.monotonic(),
                                0,
                            )
                        )
                lock_stack.enter_context(RateLimiter._lock)
            # The below is equivalent to: return (yield from self.auth_flow(request))
            # but with exponential backoff, locking, and ratelimit retry added.
            # See https://peps.python.org/pep-0380/#formal-semantics for details
            flow = self.auth_flow(request)
            try:
                _yield = next(flow)
            except StopIteration as exc:
                assert exc.value is None
                return
            backoff = _backoff[:6]
            while True:
                try:
                    _sent = yield _yield
                except GeneratorExit:
                    flow.close()
                    raise
                except BaseException as exc:
                    try:
                        _yield = flow.throw(exc)
                    except StopIteration as exc:
                        assert exc.value is None
                        return
                else:
                    status = _sent.status_code
                    if status == 429:
                        retry = _get_as_int(_sent.headers, "Retry-After", 0)
                        if retry:
                            LOG.warning("Ratelimit hit, retrying in %s seconds.", retry)
                            remaining = _get_as_int(
                                _sent.headers, "Ratelimit-Remaining", 0
                            )
                            if remaining and not is_telegram_limiter:
                                # Unmarked Telegram API
                                warnings.warn(
                                    "Telegram API was used without using sans.TelegramLimiter.",
                                    stacklevel=5,
                                )
                                # Mark this as the Telegram API for next go around
                                is_telegram_limiter = True
                                # infer the recruitment delay
                                before = time.monotonic()
                                lock_stack.close()  # let other requests go through
                                lock_stack.enter_context(
                                    TelegramLimiter._lock
                                )  # while acquiring the telegram lock in the outer stack
                                if (
                                    TelegramLimiter._last_request
                                    and TelegramLimiter._last_request > before
                                ):
                                    # another task made a TG request while we were waiting
                                    # guess when we can make a request next
                                    if retry > 30:
                                        retry = (
                                            TelegramLimiter._last_request
                                            + 180
                                            - time.monotonic()
                                        )
                                    else:
                                        retry = (
                                            TelegramLimiter._last_request
                                            + 30
                                            - time.monotonic()
                                        )
                                else:
                                    retry = before + retry - time.monotonic()
                            time.sleep(max(retry, 0))
                            lock_stack.enter_context(
                                RateLimiter._lock
                            )  # re-enter the API lock
                            backoff = _backoff[:6]
                            continue
                    elif status in [500, 502]:
                        try:
                            retry = next(backoff)
                        except StopIteration:
                            pass  # out of retries
                        else:
                            LOG.debug(
                                "Status %s received, retrying in %.0f seconds.",
                                status,
                                retry,
                            )
                            time.sleep(retry)
                            continue  # retry with the same request
                    try:
                        _yield = flow.send(_sent)
                    except StopIteration as exc:
                        if is_telegram_limiter:
                            TelegramLimiter._last_request = time.monotonic()
                        assert exc.value is None
                        return
                    backoff = _backoff[:6]

    async def async_auth_flow(
        self, request: httpx.Request
    ) -> AsyncGenerator[httpx.Request, httpx.Response]:
        recruitment_delay: int = getattr(self, "_recruitment_delay", 0)
        is_telegram_limiter = recruitment_delay > 0
        async with AsyncExitStack() as lock_stack:
            if request.headers["Host"] == API_URL.netloc.decode(
                request.headers.encoding
            ):
                if is_telegram_limiter:
                    await lock_stack.enter_async_context(TelegramLimiter._lock)
                    if TelegramLimiter._last_request is not None:
                        await anyio.sleep(
                            max(
                                TelegramLimiter._last_request
                                + recruitment_delay
                                - time.monotonic(),
                                0,
                            )
                        )
                await lock_stack.enter_async_context(RateLimiter._lock)
            # The below is equivalent to: return (yield from self.auth_flow(request))
            # but with exponential backoff, locking, and ratelimit retry added.
            # See https://peps.python.org/pep-0380/#formal-semantics for details
            flow = self.auth_flow(request)
            try:
                _yield = next(flow)
            except StopIteration as exc:
                assert exc.value is None
                return
            backoff = _backoff[:6]
            while True:
                try:
                    _sent = yield _yield
                except GeneratorExit:
                    flow.close()
                    raise
                except BaseException as exc:
                    try:
                        _yield = flow.throw(exc)
                    except StopIteration as exc:
                        assert exc.value is None
                        return
                else:
                    status = _sent.status_code
                    if status == 429:
                        retry = _get_as_int(_sent.headers, "Retry-After", 0)
                        if retry:
                            LOG.info("Ratelimit hit, retrying in %s seconds.", retry)
                            remaining = _get_as_int(
                                _sent.headers, "Ratelimit-Remaining", 0
                            )
                            if remaining and not is_telegram_limiter:
                                # Unmarked Telegram API
                                warnings.warn(
                                    "Telegram API was used without using sans.TelegramLimiter.",
                                    stacklevel=5,
                                )
                                # Mark this as the Telegram API for next go around
                                is_telegram_limiter = True
                                # infer the recruitment delay
                                before = time.monotonic()
                                await (
                                    lock_stack.aclose()
                                )  # let other requests go through
                                await lock_stack.enter_async_context(
                                    TelegramLimiter._lock
                                )  # while acquiring the telegram lock in the outer stack
                                if (
                                    TelegramLimiter._last_request
                                    and TelegramLimiter._last_request > before
                                ):
                                    # another task made a TG request while we were waiting
                                    # guess when we can make a request next
                                    if retry > 30:
                                        retry = (
                                            TelegramLimiter._last_request
                                            + 180
                                            - time.monotonic()
                                        )
                                    else:
                                        retry = (
                                            TelegramLimiter._last_request
                                            + 30
                                            - time.monotonic()
                                        )
                                else:
                                    retry = before + retry - time.monotonic()
                            await anyio.sleep(max(retry, 0))
                            await lock_stack.enter_async_context(
                                RateLimiter._lock
                            )  # re-enter the API lock
                            backoff = _backoff[:6]
                            continue
                    elif status in [500, 502]:
                        try:
                            retry = next(backoff)
                        except StopIteration:
                            pass  # out of retries
                        else:
                            LOG.debug(
                                "Status %s received, retrying in %.0f seconds.",
                                status,
                                retry,
                            )
                            await anyio.sleep(retry)
                            continue  # retry with the same request
                    try:
                        _yield = flow.send(_sent)
                    except StopIteration as exc:
                        if is_telegram_limiter:
                            TelegramLimiter._last_request = time.monotonic()
                        assert exc.value is None
                        return
                    backoff = _backoff[:6]


class TelegramLimiter(RateLimiter):
    _lock: ClassVar = ResetLock()
    _last_request: ClassVar[float] = time.monotonic() - 180

    def __init__(self, *, recruitment: bool) -> None:
        super().__init__()
        self._recruitment_delay = 180 if recruitment else 30



# MIT License

# Copyright (c) 2018 - 2025

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import sys as _sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Literal

from ._auth import *
from ._client import *
from ._errors import *
from ._limiter import RateLimiter as RateLimiter, TelegramLimiter as TelegramLimiter
from ._response import *
from ._sse import *
from ._url import *
from ._utils import *

if _sys.version_info < (3, 8):
    from importlib_metadata import metadata as _metadata
else:
    from importlib.metadata import metadata as _metadata

__copyright__ = "Copyright 2019-2025 Zephyrkul"

_meta = _metadata(__name__)
__title__ = _meta["Name"]
__author__ = _meta["Author"]
__license__ = _meta["License"]
__version__ = _meta["Version"]


del annotations, _sys, TYPE_CHECKING, _metadata, _meta  # not for export


def set_agent(new_agent: str, *, _force: Literal[False] = False) -> str:
    """
    Sets sans' User-Agent header.

    Parameters
    ----------
    agent: str
        The User-Agent header to use.
        Your nation name and a method of contact, like an email address, are recommended.
        Some script info will be appended automatically.
    _force: bool
        Forcibly override the User-Agent header even after it's already been set.
        Not recommended: https://www.nationstates.net/pages/api.html#terms

    Returns
    -------
    The actual newly set agent, including attached additional script information.
    """

    if RateLimiter._agent and not _force:
        raise RuntimeError("Agent cannot be re-set")
    if not new_agent:
        RateLimiter._agent = ""
    else:
        import sys

        import httpx

        RateLimiter._agent = (
            f"{new_agent} Python/{sys.version_info[0]}.{sys.version_info[1]} "
            f"httpx/{httpx.__version__} sans/{__version__}"
        )
    return RateLimiter._agent


def deferred() -> float | None:
    """
    Returns when the internal ratelimiting lock is scheduled to be
    released in seconds from now, or None if the lock isn't deferred.

    Note that this is only the *schedule* - the value could be
    0.0 or even negative if the deferring thread falls behind.
    """
    return RateLimiter._lock.deferred


def locked() -> bool:
    """
    Returns when the internal ratelimiting lock is locked.
    """
    return RateLimiter._lock.locked()


for obj in list(globals().values()):
    if getattr(obj, "__module__", "").startswith(__name__):
        obj.__module__ = __name__
del obj  # type: ignore



#!/usr/bin/python3

from __future__ import annotations

import argparse
import logging
import shlex
import sys
from contextlib import redirect_stdout
from getpass import getpass
from typing import TYPE_CHECKING, Any, NoReturn as Never
from xml.etree import ElementTree as ET

import sans

try:
    import readline  # noqa: F401
except ImportError:
    pass


def Syntax(arg: str, *_args: Any, **_kwargs: Any) -> object:
    return arg


if not TYPE_CHECKING:
    try:
        from rich import print
        from rich.console import Console
        from rich.logging import RichHandler
        from rich.syntax import Syntax  # noqa: F811
    except ImportError:
        logging.basicConfig(level=logging.ERROR)
    else:
        _console = Console(stderr=True)
        input = _console.input
        logging.basicConfig(
            level=logging.ERROR,
            handlers=[RichHandler(console=_console, rich_tracebacks=True)],
        )
        del _console

MAIN_LOG = logging.getLogger("sans.__main__")
SANS_LOG = logging.getLogger("sans")
ROOT_LOG = logging.getLogger()


class _ReInput:
    def __init__(self, parser: argparse.ArgumentParser) -> None:
        self.parser = parser
        self.input: list[str] | None = None

    def __enter__(self):
        return self.parser.parse_known_args(self.input)

    def __exit__(self, *exc_details):
        exc: BaseException | None = exc_details[1]
        if exc is None or isinstance(exc, Exception):
            if exc is not None:
                MAIN_LOG.exception(
                    "An exception occurred while handling your request:", exc_info=exc
                )
            try:
                with redirect_stdout(sys.stderr):
                    self.input = shlex.split(input(f"\n>>> {self.parser.prog} "))
                if not self.input:
                    sys.exit(0)
            except EOFError:
                sys.exit(0)
            return True
        return False


def main() -> Never:
    parser = argparse.ArgumentParser(
        description="SANS Console Entry",
        prog=sans.__name__,
        epilog="Any unknown args will be used to build the API request.",
    )
    parser.add_argument("--agent", "-A", help="set the script's user agent")
    parser.add_argument(
        "--auth",
        action="store_true",
        help="Prompt for a password for using private shards",
    )
    parser.add_argument(
        "--quit",
        "--exit",
        action="store_true",
        help="quit the console program loop after this run",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        "--url",
        "--headers",
        action="count",
        default=0,
        help="prints extra info about the request and response",
    )
    parser.add_argument(
        "--version", action="version", version=f"{parser.prog} {sans.__version__}"
    )
    reinput = _ReInput(parser)
    auth = sans.NSAuth()  # type: ignore
    with sans.Client() as client:
        while True:
            with reinput as (known, unknown):
                _run_once(client, auth, known, unknown)


def _run_once(
    client: sans.Client,
    auth: sans.NSAuth,
    known: argparse.Namespace,
    unknown: list[str],
) -> None:
    level = max(logging.DEBUG, logging.ERROR - logging.DEBUG * known.verbose)
    SANS_LOG.setLevel(level)
    ROOT_LOG.setLevel(level + logging.DEBUG)
    if known.auth:
        auth.autologin = None
        auth.password = getpass()
    if known.agent:
        try:
            agent = sans.set_agent(known.agent)
            print(f"Agent set: {agent}", file=sys.stderr)
        except RuntimeError:
            print(
                "You can't change the agent in the middle of the script.",
                file=sys.stderr,
            )
    if not unknown:
        if known.quit:
            print("No query provided. Exiting...", file=sys.stderr)
            sys.exit(0)
        print("No query provided.", file=sys.stderr)
        return
    parameters: dict[str, list[str]] = {}
    key = ""
    for arg in unknown:
        if arg.startswith("--"):
            if key:
                print(
                    f"No value provided for key {key!r}",
                    file=sys.stderr,
                )
            key = arg[2:]
        else:
            parameters.setdefault(key or "q", []).append(arg)
            key = ""
    if key:
        print(f"No value provided for key {key!r}", file=sys.stderr)
    request = client.build_request(
        "GET", sans.World(**{k: " ".join(v) for k, v in parameters.items()})
    )
    response = client.send(request, auth=auth, stream=True)
    try:
        response.read()
    finally:
        response.close()
    if response.content_type == "text/xml":
        pretty_print(response.xml)
    elif response.content_type.startswith("text/"):
        print(
            Syntax(
                response.text,
                response.content_type.partition("/")[2],
                background_color="default",
                word_wrap=True,
            )
        )
    else:
        print(response.content)
    if known.quit:
        print("Exiting...", file=sys.stderr)
        sys.exit(0)


def pretty_print(element: ET.Element, *, space: str = "  ") -> None:
    sans.indent(element, space)
    print(
        Syntax(
            ET.tostring(element, encoding="unicode").strip(),
            "xml",
            background_color="default",
            indent_guides=True,
            tab_size=len(space),
            word_wrap=True,
        )
    )


if __name__ == "__main__":
    logging.captureWarnings(True)
    sys.exit(main())



from __future__ import annotations

from functools import wraps
from operator import methodcaller
from typing import TYPE_CHECKING, Callable, TypeVar

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    _P = ParamSpec("_P")
    _R = TypeVar("_R")

import httpx

from ._limiter import RateLimiter

__all__ = [
    "Client",
    "ClientType",
    "AsyncClient",
    "AsyncClientType",
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
    "request",
    "stream",
]
_limiter = RateLimiter()


def _ensure_auth(
    wrapped: Callable[_P, _R], *, force_auth: bool = False
) -> Callable[_P, _R]:
    setter = methodcaller(
        "__setitem__" if force_auth else "setdefault", "auth", _limiter
    )

    @wraps(wrapped)
    def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
        setter(kwargs)
        return wrapped(*args, **kwargs)

    return inner


class Client(httpx.Client):
    __doc__ = httpx.Client.__doc__
    __init__ = _ensure_auth(httpx.Client.__init__, force_auth=True)


class AsyncClient(httpx.AsyncClient):
    __doc__ = httpx.AsyncClient.__doc__
    __init__ = _ensure_auth(httpx.AsyncClient.__init__, force_auth=True)


# backwards compatibility
ClientType = Client
AsyncClientType = AsyncClient

request = _ensure_auth(httpx.request)
delete = _ensure_auth(httpx.delete)
get = _ensure_auth(httpx.get)
head = _ensure_auth(httpx.head)
options = _ensure_auth(httpx.options)
patch = _ensure_auth(httpx.patch)
post = _ensure_auth(httpx.post)
put = _ensure_auth(httpx.put)
stream = _ensure_auth(httpx.stream)



from __future__ import annotations

import sys
from contextlib import ExitStack
from typing import Any, Coroutine, overload
from xml.etree import ElementTree as etree

import httpx

from ._auth import NSAuth
from ._client import AsyncClient, Client
from ._errors import PrivateCommandError
from ._response import Response
from ._url import Command

__all__ = ["prepare_and_execute", "indent"]


@overload
def prepare_and_execute(
    client: Client | None, auth: NSAuth, nation: str, c: str, **parameters: str
) -> Response: ...


@overload
async def prepare_and_execute(
    client: AsyncClient, auth: NSAuth, nation: str, c: str, **parameters: str
) -> Response: ...


def prepare_and_execute(
    client: Client | AsyncClient | None,
    auth: NSAuth,
    nation: str,
    c: str,
    **parameters: str,
) -> Response | Coroutine[Any, Any, Response]:
    if isinstance(client, httpx.AsyncClient):
        return _prepare_async(client, auth, nation, c, **parameters)
    with ExitStack() as stack:
        if not client:
            client = stack.enter_context(Client())
        parameters.update(mode="prepare")
        url = Command(nation, c, **parameters)
        response = client.get(url, auth=auth)
        response.raise_for_status()
        error = response.xml.findtext("ERROR", default="")
        if error:
            raise PrivateCommandError(error, response=response)
        token = response.xml.findtext("SUCCESS", default="")
        parameters.update(mode="execute", token=token)
        url = Command(nation, c, **parameters, mode="execute", token=token)
        return client.get(url, auth=auth)


async def _prepare_async(
    client: AsyncClient, auth: NSAuth, nation: str, c: str, **parameters: str
) -> Response:
    parameters.update(mode="prepare")
    url = Command(nation, c, **parameters)
    response = await client.get(url, auth=auth)
    response.raise_for_status()
    error = response.xml.findtext("ERROR", default="")
    if error:
        raise PrivateCommandError(error, response=response)
    token = response.xml.findtext("SUCCESS", default="")
    parameters.update(mode="execute", token=token)
    url = Command(nation, c, **parameters)
    return await client.get(url, auth=auth)


if sys.version_info < (3, 9):
    # from: https://github.com/python/cpython/blob/3.11/Lib/xml/etree/ElementTree.py#L1154
    def indent(
        tree: etree.Element | etree.ElementTree, space: str = "  ", level: int = 0
    ):
        if isinstance(tree, etree.ElementTree):
            tree = tree.getroot()  # type: ignore
        assert isinstance(tree, etree.Element)
        if level < 0:
            raise ValueError(f"Initial indentation level must be >= 0, got {level}")
        if not len(tree):
            return

        # Reduce the memory consumption by reusing indentation strings.
        indentations = ["\n" + level * space]

        def _indent_children(elem: etree.Element, level: int):
            # Start a new indentation level for the first child.
            child_level = level + 1
            try:
                child_indentation = indentations[child_level]
            except IndexError:
                child_indentation = indentations[level] + space
                indentations.append(child_indentation)

            if not elem.text or not elem.text.strip():
                elem.text = child_indentation

            for child in elem:
                if len(child):
                    _indent_children(child, child_level)
                if not child.tail or not child.tail.strip():
                    child.tail = child_indentation

            # Dedent after the last child by overwriting the previous indentation.
            if not child.tail.strip():  # type: ignore
                child.tail = indentations[level]  # type: ignore

        _indent_children(tree, 0)

else:
    from xml.etree.ElementTree import indent



from __future__ import annotations

from datetime import date as _date
from typing import TYPE_CHECKING, Iterable, Mapping, NewType, overload

import httpx

if TYPE_CHECKING:
    from typing_extensions import Literal, TypeAlias

    _Shard = NewType("_Shard", dict[str, str])
    _QueryValueTypes: TypeAlias = "str | int | float | bool"

API_URL = httpx.URL("https://www.nationstates.net/cgi-bin/api.cgi")

__all__ = [
    "API_URL",
    "Nation",
    "Region",
    "World",
    "WA",
    "Command",
    "Shard",
    "View",
    "Range",
    "NationsDump",
    "RegionsDump",
    "CardsDump",
    "Telegram",
]


def _primitive_value_to_str(value: object) -> str:
    # match behavior with httpx
    if value is True:
        return "true"
    elif value is False:
        return "false"
    elif value is None:
        return ""
    return str(value)


def Nation(
    nation: str, *shards: str | _Shard, **parameters: _QueryValueTypes
) -> httpx.URL:
    return World(*shards, nation=nation, **parameters)


def Region(
    region: str, *shards: str | _Shard, **parameters: _QueryValueTypes
) -> httpx.URL:
    return World(*shards, region=region, **parameters)


def World(*shards: str | _Shard, **parameters: _QueryValueTypes) -> httpx.URL:
    q: list[_QueryValueTypes | None] = [parameters.pop("q", None)]
    query: dict[str, _QueryValueTypes] = {}
    for shard in shards:
        if isinstance(shard, Mapping):
            shard = dict(shard)
            q.append(shard.pop("q", None))
            query.update(shard.items())
        else:
            q.append(shard)
    q_str = " ".join(map(_primitive_value_to_str, filter(None, q)))
    if q_str:
        query["q"] = q_str
    query.update(parameters)
    return API_URL.copy_with(params=query)


def WA(
    wa: Literal[1, "1", 2, "2"], *shards: str | _Shard, **parameters: _QueryValueTypes
) -> httpx.URL:
    return World(*shards, wa=_primitive_value_to_str(wa), **parameters)


def Command(nation: str, c: str, **parameters: _QueryValueTypes) -> httpx.URL:
    return World(nation=nation, c=c, **parameters)


def Telegram(client: str, tgid: str, key: str, to: str) -> httpx.URL:
    return World(a="sendtg", client=client, tgid=tgid, key=key, to=to)


def Shard(q: str, **parameters: _QueryValueTypes) -> _Shard:
    parameters["q"] = q
    return parameters  # type: ignore


@overload
def View(*, nations: Iterable[str]) -> _Shard: ...
@overload
def View(*, regions: Iterable[str]) -> _Shard: ...


def View(*, nations: Iterable[str] = (), regions: Iterable[str] = ()) -> _Shard:
    if bool(nations) is bool(regions):
        raise TypeError("View requires exactly one of either nations or regions")
    if nations:
        nations = "nation:" + ",".join(
            (nations,) if isinstance(nations, str) else nations
        )
        return {"view": nations}  # type: ignore
    if regions:
        regions = "region:" + ",".join(
            (regions,) if isinstance(regions, str) else regions
        )
        return {"view": regions}  # type: ignore

    raise AssertionError("Unreachable code")


def Range(__from: str | int, __to: str | int) -> _Shard:
    return {
        "from": _primitive_value_to_str(__from),
        "to": _primitive_value_to_str(__to),
    }  # type: ignore


# https://www.nationstates.net/archive/nations/2018-09-30-nations-xml.gz
def NationsDump(date: _date | None = None) -> httpx.URL:
    if date:
        path = date.strftime("/archive/nations/%Y-%m-%d-nations-xml.gz")
    else:
        path = "/pages/nations.xml.gz"
    return API_URL.join(path)


# https://www.nationstates.net/archive/nations/2018-09-30-regions-xml.gz
def RegionsDump(date: _date | None = None) -> httpx.URL:
    if date:
        path = date.strftime("/archive/nations/%Y-%m-%d-regions-xml.gz")
    else:
        path = "/pages/regions.xml.gz"
    return API_URL.join(path)


def CardsDump(season: Literal[1, "1", 2, "2", 3, "3", 4, "4"]) -> httpx.URL:
    return API_URL.join(f"/pages/cardlist_S{season}.xml.gz")



from __future__ import annotations

import asyncio
import threading
import time
import weakref
from collections import deque
from contextvars import Context, copy_context
from functools import partial
from operator import add
from typing import TYPE_CHECKING, Any, Awaitable, Callable

import anyio
import sniffio
from anyio.lowlevel import cancel_shielded_checkpoint, checkpoint_if_cancelled

if TYPE_CHECKING:
    from typing_extensions import TypeVarTuple, Unpack

    _Ts = TypeVarTuple("_Ts")

__all__ = ["ResetLock"]


def _threading_sleep_forever(callback_deque: deque[Callable[[], Any]]):
    lock = threading.Lock()
    with lock:
        callback_deque.append(lock.release)
        try:
            lock.acquire()
        finally:
            callback_deque.remove(lock.release)


async def _asyncio_sleep_forever(callback_deque: deque[Callable[[], Any]]) -> None:
    future = asyncio.get_running_loop().create_future()
    release = partial(future.get_loop().call_soon_threadsafe, future.set_result, None)
    callback_deque.append(release)  # type: ignore
    try:
        await future
    finally:
        callback_deque.remove(release)  # type: ignore


try:
    import trio
except ImportError:
    pass
else:

    @trio.lowlevel.enable_ki_protection
    async def _trio_sleep_forever(
        callback_deque: deque[Callable[[], Any]],
    ) -> Awaitable[None]:
        task = trio.lowlevel.current_task()
        release = partial(
            trio.lowlevel.current_trio_token().run_sync_soon,
            trio.lowlevel.reschedule,
            task,
        )
        callback_deque.append(release)  # type: ignore

        try:
            return await trio.lowlevel.wait_task_rescheduled(
                lambda _: trio.lowlevel.Abort.SUCCEEDED
            )
        finally:
            callback_deque.remove(release)  # type: ignore


class _ThreadTimer(threading.Timer):
    def __init__(
        self,
        delay: float,
        callback: Callable[[Unpack[_Ts]], Any],
        *args: Unpack[_Ts],
        context: Context | None = None,
    ) -> None:
        if not context:
            context = copy_context()
        super().__init__(delay, context.run, (callback, *args))

    def when(self) -> float:
        # when the timer completes if started at this very moment
        return self.interval + time.monotonic()

    def run(self) -> None:
        # freeze .when() at the current time
        self.when = partial(add, self.interval, time.monotonic())
        try:
            return super().run()
        finally:
            self.when = time.monotonic
            del self.function, self.args, self.kwargs


# I would love to make this utilize a semaphore instead
# However, since multiple clients can exist on one machine,
# locking and checking the headers sequentially is a requirement
class ResetLock:
    """
    Combination :class:`asyncio.Lock`, :class:`trio.Lock`, and :class:`threading.Lock`
    for ratelimiting purposes. Works based on a FIFO queue.

    Usage::

        lock = ResetLock()
        [async] with lock.maybe_lock(url):
            response = ...
            lock.maybe_defer(response.headers)

    Note that, when using :mod:`httpx`,
    using :class:`RateLimiter` will manage a global lock for you.
    """

    __deferred: _ThreadTimer | None = None

    def __finalizer(self) -> None:
        pass

    def __init__(self):
        self._lock = threading.Lock()
        self._waiters: deque[Callable[[], Any]] = deque()

    def _wake_up_next(self):
        try:
            release = next(iter(self._waiters))
        except StopIteration:
            return
        else:
            release()

    async def __aenter__(self) -> None:
        acquire = self._lock.acquire
        waiters = self._waiters
        sleep_forever: Callable[[deque[Callable[[], Any]]], Awaitable[None]] = (
            globals()[f"_{sniffio.current_async_library()}_sleep_forever"]
        )
        await checkpoint_if_cancelled()
        if not waiters:
            if acquire(False):
                await cancel_shielded_checkpoint()
                return
        try:
            await sleep_forever(waiters)
        except anyio.get_cancelled_exc_class():
            if not self._lock.locked():
                self._wake_up_next()
            raise
        else:
            # We were just rescheduled, so the lock is about to be released
            acquire()

    async def __aexit__(self, *_):
        self.release()

    def __enter__(self) -> bool:
        acquire = self._lock.acquire
        waiters = self._waiters
        if not waiters:
            if acquire(False):
                return True
        _threading_sleep_forever(waiters)
        return acquire()

    def __exit__(self, *_):
        self.release()

    @property
    def deferred(self) -> float | None:
        if self.__deferred:
            return time.monotonic() - self.__deferred.when()
        # return None

    _deferred = property(deferred.fget)

    @_deferred.setter
    def _deferred(self, value: _ThreadTimer):
        assert self.locked()
        self.__finalizer()
        value.start()
        self.__deferred = value
        self.__finalizer = weakref.finalize(self, value.finished.set)

    @_deferred.deleter
    def _deferred(self) -> None:
        assert self.locked()
        self.__finalizer()
        del self.__deferred

    def defer(self, delay: float) -> None:
        if not self.locked():
            return
        self._deferred = _ThreadTimer(delay, self._release)

    def locked(self):
        return self._lock.locked()

    def release(self):
        if self.deferred is not None:
            return
        self._wake_up_next()
        self._lock.release()

    def _release(self):
        del self._deferred
        self._wake_up_next()
        self._lock.release()


if __name__ == "__main__":
    from concurrent.futures import ThreadPoolExecutor

    from anyio.lowlevel import checkpoint

    async def main():
        lock = ResetLock()

        def target(lock=lock):
            print("before threading lock")
            with lock:
                print("during threading lock")
                lock.defer(1)
            print("after threading lock")

        async def async_target(lock=lock):
            lib = sniffio.current_async_library()
            print(f"before {lib} lock")
            async with lock:
                print(f"during {lib} lock")
                lock.defer(1)
                await checkpoint()
            print(f"after {lib} lock")

        with ThreadPoolExecutor() as pool:
            async with anyio.create_task_group() as group:
                for _ in range(5):
                    group.start_soon(async_target)
                    pool.submit(target)
                    await checkpoint()

    anyio.run(main, backend="asyncio")
    anyio.run(main, backend="trio")



from __future__ import annotations

from datetime import datetime, timezone
from itertools import chain
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Generic,
    Iterable,
    Iterator,
    TypedDict,
    TypeVar,
)
from urllib.parse import quote

import httpx

from ._client import AsyncClient, Client
from ._eventsource import Event, EventSource
from ._url import API_URL

if TYPE_CHECKING:
    from json import loads
else:
    try:
        from orjson import loads
    except ModuleNotFoundError:
        from json import loads

__all__ = ["serversent_events"]
_T = TypeVar("_T")
_ClientT = TypeVar("_ClientT", Client, AsyncClient)
_OptionalClientT = TypeVar("_OptionalClientT", Client, AsyncClient, None)


# backport of 3.10's contextlib.nullcontext
class _nullcontext(Generic[_T]):
    __slots__ = ("enter_result",)

    def __init__(self, enter_result: _T):
        self.enter_result = enter_result

    def __enter__(self) -> _T:
        return self.enter_result

    def __exit__(self, *_):
        pass

    async def __aenter__(self) -> _T:
        return self.enter_result

    async def __aexit__(self, *_):
        pass


class _SSEvent(TypedDict):
    str: str
    htmlStr: str
    id: int
    time: datetime
    buckets: list[str]


def _decode_event_data(event: Event) -> _SSEvent:
    data = loads(event.data)
    data["id"] = int(data["id"])
    data["time"] = datetime.fromtimestamp(data["time"], tz=timezone.utc)
    return data


def _wrap_client(client: _ClientT | None) -> _nullcontext[_ClientT] | None:
    if client is None:
        return None
    return _nullcontext(client)


def _quote_parameters(
    buckets: Iterable[str], nations: Iterable[str], regions: Iterable[str]
) -> str:
    return quote(
        "+".join(
            chain(
                buckets,
                nations
                and map(
                    "nation:%s".__mod__,
                    (nations,) if isinstance(nations, str) else nations,
                ),
                regions
                and map(
                    "region:%s".__mod__,
                    (regions,) if isinstance(regions, str) else regions,
                ),
            )
        ),
        safe="+: ",
    ).replace(" ", "_")


class _SSViewIter(Generic[_OptionalClientT]):
    __slots__ = ("_client", "_url")

    def __init__(self, client: _OptionalClientT, url: httpx.URL):
        self._client = client
        self._url = url

    def __iter__(self: _SSViewIter[Client] | _SSViewIter[None]) -> Iterator[_SSEvent]:
        url = self._url
        with _wrap_client(self._client) or Client() as client:
            del self
            yield from map(_decode_event_data, EventSource(client, url=url))

    async def __aiter__(
        self: _SSViewIter[AsyncClient] | _SSViewIter[None],
    ) -> AsyncIterator[_SSEvent]:
        url = self._url
        async with _wrap_client(self._client) or AsyncClient() as client:
            del self
            async for event in EventSource(client, url=url):
                yield _decode_event_data(event)

    def __repr__(self):
        return f"<{self.__class__.__name__} client={self._client!r} url={self._url!r}"


class _SSIter(_SSViewIter[_OptionalClientT]):
    __slots__ = ()

    def view(
        self, *views: str, nations: Iterable[str] = (), regions: Iterable[str] = ()
    ) -> _SSViewIter[_OptionalClientT]:
        """
        Subscribe only to events that also match at least one of the provided buckets.

        Parameters
        ----------
        views: *str
            The different buckets to filter with.
        nations: Iterable[str]
            The different nation-based buckets to also filter with.
        regions: Iterable[str]
            The different region-based buckets to also filter with.

        Returns
        -------
        An optionally asynchronous iterable which, when iterated,
        yields events as they occur in the form of a dict.
        """
        quoted = _quote_parameters(views, nations, regions)
        if not quoted:
            raise TypeError("At least one view is required.")
        url = API_URL.copy_with(
            raw_path=b"/".join((self._url.raw_path, quoted.encode("ascii")))
        )
        return _SSViewIter(self._client, url)


def serversent_events(
    client: _OptionalClientT,
    *buckets: str,
    nations: Iterable[str] = (),
    regions: Iterable[str] = (),
) -> _SSIter[_OptionalClientT]:
    """
    Subscribe to and iterate over server-sent events.

    Parameters
    ----------
    client: Client | AsyncClient | None
        The client to use. If not supplied,
        one will be opened for the lifetime of the iterator.
    buckets: *str
        The different buckets to subscribe to.
    nations: Iterable[str]
        The different nation-based buckets to also subscribe to.
    regions: Iterable[str]
        The different region-based buckets to also subscribe to.

    Returns
    -------
    An optionally asynchronous iterable which, when iterated,
    yields events as they occur in the form of a dict.
    """
    quoted = _quote_parameters(buckets, nations, regions)
    if not quoted:
        raise ValueError("At least one bucket is required.")
    # use raw_path or httpx will do its own standards-compliant encoding
    url = API_URL.copy_with(raw_path=b"/api/" + quoted.encode("ascii"))
    return _SSIter(client, url)
